#' Gets nice break points & labels for plotting contours
#'
#' \code{csteps} gets nice equally-spaced break points for plotting contours, based
#'    on the range provided, and generates labels
#'
#' @param min A numeric value. The lower bound of the range.
#' @param max A numeric value. The upper bound of the range.
#' @param nsteps A numeric value. The target number of categories required.
#' @param ticks A numeric vector. A list of values suitable for
#'     tick marks (modulo powers of 10); default = \code{c(1,2,2.5,5,7.5,10)}
#'
#' @return A list of length 3:
#' \enumerate{
#' \item \code{breaks}: a numeric vector containing the set of break points (length \code{nbreaks})
#' \item \code{labels}: a character vector containing the set of labels (length \code{nbreaks-1})
#' \item \code{nbreaks}: the number of breaks (may be different from \code{nsteps})
#' }
#'
csteps <- function(min, max, nsteps, ticks=c(1,2,2.5,4,5,7.5,10)){
  range <- max - min
  step <- range/nsteps
  roundstep <- 10^(floor(log10(step)))
  ratio <- step/roundstep
  best <- min(which(ticks >= ratio))
  newstep <- ticks[[best]]*roundstep
  newmin <- floor(min/newstep)*newstep
  newmax <- ceiling(max/newstep)*newstep
  breaks <- seq(from=newmin, to=newmax, by=newstep)
  nbreaks <- length(breaks)
  scale <- abs(floor(log10(newstep)))+2
  nm1 <- nbreaks-1
  labels<- paste0(signif(breaks[1:nm1],scale), "-", signif(breaks[2:nbreaks],scale))
  res <- list()
  res$breaks <- breaks
  res$labels <- labels
  res$nstep <- nbreaks
  res
} # end of function
