#'
#' Calculates EDF for set of random terms in asreml object
#'
#' \code{edf} takes an asreml object and calculates EDF (effective degrees of freedom,
#'  cf Lee, Durban & Eilers, 2013) for all random terms in the fitted model.
#'
#' Warnings: \itemize{
#' \item in the current version, this function only works when all terms have (scaled) identity variance models.
#' If there are non-identity variance models present, an error is generated.
#' \item the estimate may be inaccurate if convergence has not been achieved.
#' Eg. use \code{asreml.options(extra=10)} to force 10 extra iterations. If convergence has not
#' been achieved and the variance component is small, negative edf may occur.
#' }
#'
#' @param asr.object An asreml object from a fitted model
#'
#' @return A data frame with 2 columns: the term names and
#'     their calculated edf.
#'
#' @examples
#' \dontrun{
#' data("wheatdata")
#' wheatdata$R <- as.factor(wheatdata$row)
#' wheatdata$C <- as.factor(wheatdata$col)
#'
#' # Tensor-product spline using mbf in the asreml model
#' TPXZ <- tpsmmb("col", "row", wheatdata, stub="1", nsegments=c(14,21))
#' BcZ1.df <- TPXZ$BcZ.df
#' BrZ1.df <- TPXZ$BrZ.df
#' BcrZ1.df <- TPXZ$BcrZ.df
#' wh1.asr <- asreml(yield~1+TP.CR.2+TP.CR.3+TP.CR.4+colcode+rowcode+geno,
#'                  random=~R + C + TP.C.1:mbf(TP.row) + TP.C.2:mbf(TP.row) +
#'                          TP.R.1:mbf(TP.col) + TP.R.2:mbf(TP.col) +
#'                          mbf(TP.CxR),
#'                  mbf=TPXZ$mbflist,
#'                  data=TPXZ$data)
#' edf(wh1.asr)
#' tpsfit1 <- tpsfitted(wh1.asr, TPXZ)
#' tpsurface(tpsfit1)
#' }
#'
#' @export

edf <-
  function(asr.object)  {

# check for non-id variance models
# asr.object$G.param$term$factor$struc - 3rd element indicates variance model
  nterms <- length(asr.object$G.param)
  nvarmod <- 0
  for (i in 1:nterms) {
    nfac <- length(asr.object$G.param[[i]])
    for (j in 2:nfac) {
      k <- match("struc",names(asr.object$G.param[[i]][[j]]))
      if (asr.object$G.param[[i]][[j]][[k]][[3]]>0) {
      # Found non-ID variance model
        stop("edf calculations wrong when variance model present")
      }
    }
  }

# get variances of coefficients for random terms
  diagC <- asr.object$vcoeff$random
# get set of coefficients
  coef <- asr.object$coefficients
# get pointers to individual random terms
  tlev <-  attributes(coef$random)$terms$n
  tnames <- attributes(coef$random)$terms$tname
  tcum <- cumsum(tlev)
  tstart <- tidyr::replace_na(dplyr::lag(tcum,1),0)+1
  tend <- tcum

# calculate EDF
# get number of terms
  nrt <- length(tnames)
# get gammas
  gammas <- asr.object$vparameters
# initialise EDF vector
  EDF <- vector()
  length(EDF) <- nrt
# calculate edf for each term - assumes identiy variance model
  for (i in 1:nrt) {
    Cinv.term <- diagC[c(tstart[i]:tend[i])]
    EDF[i] <- round(length(Cinv.term) - sum(Cinv.term)/gammas[i],digits=3)
  }
# put edf into data frame with term names & return
  edf.df <- data.frame(tnames,EDF)
  res <- edf.df
  res
}
