#'
#' Get components of fitted tensor-product spline
#'
#' \code{tpsfitted} gets components of fitted tensor product spline from an asreml object and
#'     the object generated by the \code{tpsmmb} function used to generate the spline design
#'     matrices
#'
#' Assumes the the spline design
#'     matrices generated by the \code{tpsmmb} function are present in the model
#'     using the same names.
#'
#' @param asr.object An \code{asreml} object. The \code{asreml} object used to fit the tensor-product spline.
#' @param tps.object An object from a call to \code{tpsmmb} used in the \code{asreml} fit
#'
#' @return A data frame with fitted values for each component of the  tensor product spline and their total.
#' Note: the fixed component of the model is centered to exclude the overall constant term.
#'
#' @examples
#' \dontrun{
#' data("wheatdata")
#' wheatdata$R <- as.factor(wheatdata$row)
#' wheatdata$C <- as.factor(wheatdata$col)
#'
#' # Tensor-product spline using mbf in the asreml model
#' TPXZ <- tpsmmb("col", "row", wheatdata, stub="1", nsegments=c(14,21))
#' BcZ1.df <- TPXZ$BcZ.df
#' BrZ1.df <- TPXZ$BrZ.df
#' BcrZ1.df <- TPXZ$BcrZ.df
#' wh1.asr <- asreml(yield~1+TP.CR.2+TP.CR.3+TP.CR.4+colcode+rowcode+geno,
#'                  random=~R + C + TP.C.1:mbf(TP.row) + TP.C.2:mbf(TP.row) +
#'                          TP.R.1:mbf(TP.col) + TP.R.2:mbf(TP.col) +
#'                          mbf(TP.CxR),
#'                  mbf=TPXZ$mbflist,
#'                  data=TPXZ$data)
#' edf(wh1.asr)
#' tpsfit1 <- tpsfitted(wh1.asr, TPXZ)
#' tpsurface(tpsfit1)
#'
#' }
#'
#' @export

tpsfitted <- function(asr.object, tps.object){

# Deal with the random part of the spline

    # get pointers into the individual random terms
    coef <- asr.object$coefficients
    tlev <-  attributes(coef$random)$terms$n
    tnames <- attributes(coef$random)$terms$tname
    tcum <- cumsum(tlev)
    tstart <- tidyr::replace_na(dplyr::lag(tcum,1),0)+1
    tend <- tcum
    # find which terms are tensor-product terms (start "TP.")
    tpterms <- grep("TP.",tnames)
    nterms <- length(tpterms)

    # get coordinates vectors: short version & mapped onto full length
    col.vector <- tps.object$data[["TP.col"]]
    col.match <- match(col.vector,sort(unique(col.vector)))
    row.vector <- tps.object$data[["TP.row"]]
    row.match <- match(row.vector,sort(unique(row.vector)))
    cr.vector <- tps.object$data[["TP.CxR"]]
    cr.match <- match(sort(unique(cr.vector)),cr.vector)

    # loop over random terms and accumulate fitted values
    fitted <- list()
    names <- list()
    for (i in 1:nterms){
      # get variables in current term
      k <- tpterms[i]
      term <- tnames[k]
      f <- stats::as.formula(paste("~",term))
      v <- all.vars(f)
      nv <- length(v)
      if (nv>2){stop("Problem: wrong number of variables in random tensor-product term - should be <=2")}
      # loop over terms and extract component design matrices from tps.object
      mat <- list()
      ncol <- list()
      namestub <- list()
      # check which type of variable, extract design matrix & name
      for (j in 1:nv){
        # check for grp variable
        if (!is.null(tps.object[["grp"]])) {
          if (v[j]%in%names(tps.object$grp)) {
            index <- which(names(tps.object$grp)==v[j], arr.ind=TRUE)
            mat[[j]] <- as.matrix(tps.object$data[tps.object$grp[[index]]])
            ncol[[j]] <- length(tps.object$grp[[index]])
            namestub[[j]] <- names(tps.object$grp)[[index]]
            next()
          }
        }
        # X part for rows - full length
        if (!(length(grep("TP\\.R\\.",v[j]))==0)){
          index <- match(v[j],names(tps.object$data))
          mat[[j]] <- as.matrix(tps.object$data[[index]])
          ncol[[j]] <- 1
          nc <- nchar(v[j])
          namestub[[j]] <- paste0("row",substr(v[j],nc,nc))
          next()
        }
        # X part for columns - full length
        if (!(length(grep("TP\\.C\\.",v[j]))==0)){
          index <- match(v[j],names(tps.object$data))
          mat[[j]] <- as.matrix(tps.object$data[[index]])
          ncol[[j]] <- 1
          nc <- nchar(v[j])
          namestub[[j]] <- paste0("col",substr(v[j],nc,nc))
          next()
        }
        # Z part for rows - held short, needs expanding
        if (!(length(grep("TP\\.row",v[j]))==0)){
          index <- pmatch("BrZ",names(tps.object))
          ncol[[j]] <- length(tps.object[[index]])-1
          submat <- as.matrix(tps.object[[index]])
          mat[[j]] <- submat[row.match,c(1:ncol[[j]])]
          namestub[[j]] <- paste("f(row)")
          next()
        }
        # Z part for columns - held short, needs expanding
        if (!(length(grep("TP\\.col",v[j]))==0)){
          index <- pmatch("BcZ",names(tps.object))
          ncol[[j]] <- length(tps.object[[index]])-1
          submat <- as.matrix(tps.object[[index]])
          mat[[j]] <- submat[col.match,c(1:ncol[[j]])]
          namestub[[j]] <- paste("f(col)")
          next()
        }
        # Z part for row:col - held full length, may need reordering
        if (!(length(grep("TP\\.CxR",v[j]))==0)){
          index <- pmatch("BcrZ",names(tps.object))
          ncol[[j]] <- length(tps.object[[index]])-1
          submat <- as.matrix(tps.object[[index]])
          mat[[j]] <- submat[cr.match,c(1:ncol[[j]])]
          namestub[[j]] <- paste("f(row,col)")
          next()
        }
        # if we have got here then its not a TP name - check variate not factor
        index <- match(v[j],names(tps.object$data))
        if (is.na(index)) stop("Unknown variable in model term")
        if (is.factor(tps.object$data[[index]])) stop("Non-TP variables in TP terms must be variates")
        mat[[j]] <- as.matrix(tps.object$data[[index]])
        ncol[[j]] <- 1
        nc <- nchar(v[j])
        namestub[[j]] <- names(tps.object$data)[[index]]
      } # end of j for loop
      # construct design matrix and term name
      if(nv==2){
        mat1r <- matrix(rep(1,times=ncol[[2]]),nrow=1,ncol=ncol[[2]])
        mat1c <- matrix(rep(1,times=ncol[[1]]),nrow=1,ncol=ncol[[1]])
        desmat <- (mat[[1]]%x%mat1r)*(mat1c%x%mat[[2]])
        termname <- paste(namestub[[1]],namestub[[2]],sep=":")
      }
      else{
        desmat <- mat[[1]]
        termname <- namestub[[1]]
      }
      # get coefficients
      u <- coef$random[c(tstart[[k]]:tend[[k]])]
      fitted[[i]] <- desmat%*%u
      names[[i]] <- termname
    } # end of i for loop
    names(fitted) <- names

# Deal with the fixed part of the spline: all single column terms
    # get the fixed terms individually
    tlev <-  attributes(coef$fixed)$terms$n
    tnames <- attributes(coef$fixed)$terms$tname
    tcum <- cumsum(tlev)
    tstart <- tidyr::replace_na(dplyr::lag(tcum,1),0)+1
    tend <- tcum
    # find the tensor-product terms (start with "TP.")
    tpterms <- grep("TP.",tnames)
    ntermsf <- length(tpterms)
    # skip next section if no fixed terms
    if (ntermsf>0) {
    # loop over terms to find tensor product terms
      fitfix <- rep(0,times=length(col.vector))
      for (i in 1:ntermsf){
        # get variables in current term
        k <- tpterms[i]
        term <- tnames[k]
        f <- stats::as.formula(paste("~",term))
        v <- all.vars(f)
        nv <- length(v)
        if (nv>1){stop("Problem: wrong number of variables in fixed tensor-product term - should be ==1")}
        # get coefficients
        u <- coef$fixed[c(tstart[[k]]:tend[[k]])]
        # check which variable
        index <- match(v[1],names(tps.object$data))
        fitfix <- fitfix + tps.object$data[[index]]*u
      } # end of i for loop
      # center fixed terms to remove intercept, if present
      fitfix <- fitfix - mean(fitfix)
    }

    # save individual fitted values & total spatial surface
    fitted.df <- data.frame("col"=col.vector, "row"=row.vector)
    total <- rep(0,times=length(col.vector))
    if (ntermsf>0) {
      fitted.df$fixed <- fitfix
      total <- fitfix
    }
    for (i in 1:nterms){
      label <- names[[i]]
      fitted.df[[label]] <- as.vector(fitted[[i]])
      total <- total + fitted.df[[label]]
    } # end of second i for loop
    fitted.df$total <- total

    # return data frame of fitted value vectors
    res <- fitted.df
  } # end of function
