#'
#' Contour or tile plot of surfaces
#'
#' \code{tpsurface} produces contour (or tile) plots of components of a fitted
#'     tensor-product spline, on common or separate scales
#'
#' @param data A data frame. First two columns must be indexing vectors
#'     named \code{"col"} and \code{"row"} giving the plot locations, remaining
#'     columns are the surfaces to be plotted.
#' @param method A string. Whether to do a contour plot (interpolated)
#'     \code{"contour"} or tile plot (actual values) \code{"tile"};
#'     default \code{"contour"}.
#' @param scale A string. Whether to use a common scale across all plots
#'     \code{"common"}, or separate scales \code{"separate"};
#'     default \code{"common"}
#' @param yaxis A string. Whether to reverse y-axis so plots are in field
#'     layout with (1,1) at top left \code{"reverse"} or leave in Cartesian
#'     coordinates with (1,1) at bottom left \code{"original"}; default \code{"reverse"}.
#' @param grid A string. Whether to put all plots in grid on same page;
#'     default \code{"yes"} or use separate pages \code{"no"}; default \code{"yes"}.
#'     Note: a legend is provided only when scale=\code{"common"}.
#' @param nbreaks A numeric value. The target number of breaks to use in plotting
#'     (=number of colours used); default 12.
#' @param layout A list with 2 values: number of columns & rows in the graphical layout.
#'     Only relevant when \code{grid="yes"}. Default: determined automatically.
#'
#' @return NULL
#'
#' @examples
#' \dontrun{
#' data("wheatdata")
#' wheatdata$R <- as.factor(wheatdata$row)
#' wheatdata$C <- as.factor(wheatdata$col)
#'
#' # Tensor-product spline using mbf in the asreml model
#' TPXZ <- tpsmmb("col", "row", wheatdata, stub="1", nsegments=c(14,21))
#' BcZ1.df <- TPXZ$BcZ.df
#' BrZ1.df <- TPXZ$BrZ.df
#' BcrZ1.df <- TPXZ$BcrZ.df
#' wh1.asr <- asreml(yield~1+TP.CR.2+TP.CR.3+TP.CR.4+colcode+rowcode+geno,
#'                  random=~R + C + TP.C.1:mbf(TP.row) + TP.C.2:mbf(TP.row) +
#'                          TP.R.1:mbf(TP.col) + TP.R.2:mbf(TP.col) +
#'                          mbf(TP.CxR),
#'                  mbf=TPXZ$mbflist,
#'                  data=TPXZ$data)
#' edf(wh1.asr)
#' tpsfit1 <- tpsfitted(wh1.asr, TPXZ)
#' tpsurface(tpsfit1)
#' }
#'
#' @export

tpsurface <-
  function(data, method='contour', scale='common',yaxis='reverse',
            grid='yes', nbreaks=12, layout=NULL){

# need to accumulate plots if using grid
    if ('yes'%in%grid) P <- list()
# get number of plots - assume first two columns are col and row
    ndf <- length(data)
    nplots <- ndf-2
    plot <- list()
    nmstring <- list()
# get breaks, labels and colours if common scale to be used across plots
    if ('common'%in%scale){
      max <- max(sapply(data[-c(1,2)],max,na.rm=TRUE))
      min <- min(sapply(data[-c(1,2)],min,na.rm=TRUE))
      contours <- csteps(min,max,nbreaks)
      gg_color_hue <- function(n) {
        hues = seq(15, 300, length = n + 1)
        colors <- grDevices::hcl(h = hues, l = 65, c = 100)
        colors[1:n]
      }
    }
# start loop over plots
    for (i in 1:nplots){
      k <- i+2
      # get name of column as plot title
      nm <- names(data)
      nmstring[[i]] <- paste("`",nm[[k]],"`",sep="")
      # put together contour plot instructions with common or separate scales
      if('contour'%in%method){
        plot[[i]] <- ggplot2::ggplot(data, ggplot2::aes_string(y="row", x="col",
                                                      z=nmstring[[i]]))
        if ('common'%in%scale) {
          plot[[i]] <- plot[[i]] +
                       ggplot2::geom_contour_filled(breaks=contours$breaks)
          plot[[i]] <- plot[[i]] +
                       ggplot2::scale_fill_manual(palette=gg_color_hue,
                                                  values=contours$labels,
                                                  name="Value", drop=FALSE)
        }
        else {
          plot[[i]] <- plot[[i]] + ggplot2::geom_contour_filled()
        }
      }
      # or tile plot instructions with common or separate scales
      else{
        # need to remove missing values
        na.rm.data <- stats::na.omit(data)
        if ('common'%in%scale) {
          intfac <- cut(na.rm.data[[k]],breaks=contours$breaks)
          plot[[i]] <- ggplot2::ggplot(na.rm.data, ggplot2::aes(y=row, x=col)) +
                       ggplot2::geom_tile(ggplot2::aes_string(fill=intfac))
          plot[[i]] <- plot[[i]] +
                        ggplot2::scale_fill_manual(palette=gg_color_hue,
                                                   values=contours$labels,
                                                   name="Value", drop=FALSE)
        }
        else {
          plot[[i]] <- ggplot2::ggplot(na.rm.data, ggplot2::aes(y=row, x=col)) +
                       ggplot2::geom_tile(ggplot2::aes_string(fill=nmstring[[i]]))
          plot[[i]] <- plot[[i]] +
                       ggplot2::scale_fill_gradientn(colours=grDevices::topo.colors(nbreaks))
        }
      }
      # reverse y-axis by default
      if ('reverse'%in%yaxis)
        plot[[i]] <- plot[[i]] + ggplot2::scale_y_reverse()
      # for plots on a grid, extract legend and add plot titles
      if ('yes'%in%grid) {
        if (i==1) legend <- cowplot::get_legend(plot[[1]])
        P <- list(P, list(plot[[i]]))
        plot[[i]] <- plot[[i]] + ggplot2::theme(legend.position = "none") +
                                 ggplot2::ggtitle(nmstring[[i]])
      }
      # for plots on separate pages, just add plot titles & print plot
      else {
        print(plot[[i]]+ggplot2::ggtitle(nmstring[[i]]))
      }
    } # end loop over plots

# for grid of plots, create layout with legend down rhs
# when scale not common, shouldnt add space for legend!
    if ('yes'%in%grid) {
      np1 <- nplots+1
      if ('common'%in%scale) plot[[np1]] <- legend
      if (missing(layout)) {
        mfrow <- ceiling(sqrt(nplots))
        mfcol <- ceiling(nplots/mfrow)
      }
      else {
        mfcol <- layout[1]
        if (length(layout)>=2) {mfrow <- layout[2]}
        else {mfrow <- ceiling(nplots/mfcol)}
        if (mfrow*mfcol<nplots)
          stop("Specified layout too small for number of plots")
      }
      pos <- list()
      layoutx <- paste0("rbind(")
      for (i in 1:mfrow){
        set <- (i-1)*mfcol + c(1:mfcol)
        set <- replace(set, set>nplots, NA)
        pos[[i]] <- c(set,np1)
        layoutx <- paste0(layoutx,"pos[[",i,"]]")
        if (i<mfrow) layoutx <- paste0(layoutx,",")
      }
      layoutx <- paste0(layoutx,")")
      lay <- eval(parse(text=layoutx))
      do.call(gridExtra::grid.arrange, c(list(grobs=plot),
                                         list(layout_matrix=lay)))
    }

  invisible(NULL)
} # end of function
